/* $Header: /cvsroot/sid/sid/user/util.c,v 1.31 2004/08/09 19:23:47 deppeler Exp $ */

#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include "c_tsw.h"
#include "sha1_kernel.h"

extern struct config *config; /* global, yes */
extern int debug;	      /* global, too */

/* setprogname: set stored name of program */
void setprogname(char *str)
{
	config->name = estrdup(str);
}

/* estrdup: duplicate a string, report if error 
 * args:	*s	string to duplicate
 *
 * return values:	pointer to duplicated string */
char *estrdup(const char *s)
{
	char *t;

	t = (char *)malloc(strlen(s) + 1);
	if (NULL == t)
		eprintf("estrdup(\"%.20s\") failed:", s);
	strcpy(t, s);
	return t;
}

/* progname: return stored name of program */
char *progname(void)
{
	return config->name;
}

/* eprintf: print error message and exit */
void eprintf(char *fmt, ...)
{
	va_list args;

	fflush(stdout);
	if (progname() != NULL)
		fprintf(stderr, "%s: ", progname());

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);

	if ('\0' != fmt[0] && ':' == fmt[strlen(fmt) - 1])
		fprintf(stderr, " %s", strerror(errno));
	fprintf(stderr, "\n");
	exit(2);		/* conventional value for failed execution */
}

/* debug printing 
 * args:	l	debug level 
 * 		*fmt	format of the debug message
 * 		...	args to the format */
void dprintf(int l, char *fmt, ...)
{
	va_list args;

	if (debug < l)
		return;

	fflush(stdout);
	if (progname() != NULL)
		fprintf(stderr, "%s [debug]: ", progname());

	fprintf(stderr, "[%d] ", l);

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);

	if ('\0' != fmt[0] && ':' == fmt[strlen(fmt) - 1])
		fprintf(stderr, " %s", strerror(errno));
	fprintf(stderr, "\n");
}

/* free host_cfg list */
void free_host_cfg(struct host_cfg *c)
{
	struct host_cfg *next;
	struct tsw_time *next_automated_time;
	struct tsw_cmd *next_allowed_command_set;

	for (; c != NULL; c = next) {

		/* free automated_time list */
		for (; c->automated_time != NULL;
		     c->automated_time = next_automated_time) {
			next_automated_time = c->automated_time->next;
			free(c->automated_time);
		}
		/* free command set list */
		for (; c->allowed_command_set != NULL;
		     c->allowed_command_set = next_allowed_command_set) {
			next_allowed_command_set = c->allowed_command_set->next;
			free(c->allowed_command_set);
		}

		next = c->next;
		free(c);
	}
}

/* free cmd_def list */
void free_cmd_def(struct cmd_def *cd)
{
	struct cmd_def *next;

	for (; cd != NULL; cd = next) {
		next = cd->next;
		free(cd);
	}
}

/* free config struct */
void free_all(struct config *c)
{
	free_host_cfg(c->host);
	free_cmd_def(c->command);
	free(c->name);
	free(c);
}

/* allocate new command definition */
struct cmd_def *new_cmd_def(void)
{
	struct cmd_def *newp;

	newp = (struct cmd_def *)ecalloc(1, sizeof(struct cmd_def));
	newp->next = NULL;
	return newp;
}

/* allocate new command */
struct tsw_cmd *new_cmd(void)
{
	struct tsw_cmd *newp;

	newp = (struct tsw_cmd *)ecalloc(1, sizeof(struct tsw_cmd));
	newp->prev = NULL;
	newp->next = NULL;
	return newp;
}

/* allocate new time specification */
struct tsw_time *new_time(void)
{
	struct tsw_time *newp;

	newp = (struct tsw_time *)ecalloc(1, sizeof(struct tsw_time));
	newp->next = NULL;
	return newp;
}

/* add new command definition in front (left) of the list */
struct cmd_def *add_cmd_def_l(struct cmd_def *listp, struct cmd_def *newp)
{
	newp->next = listp;
	return newp;
}

/* add new command in front (left) of the list */
struct tsw_cmd *add_cmd_l(struct tsw_cmd *listp, struct tsw_cmd *newp)
{
	newp->next = listp;
	if (listp != NULL)
		listp->prev = newp;
	return newp;
}

/* add_time_front: add newp in fron of listp */
struct tsw_time *add_time_front(struct tsw_time *listp, struct tsw_time *newp)
{
	newp->next = listp;
	return newp;
}

/* addfront: add newp to front of listp */
struct host_cfg *addfront(struct host_cfg *listp, struct host_cfg *newp)
{
	newp->next = listp;
	return newp;
}

/* ecalloc: malloc and report if error */
void *ecalloc(size_t nelem, size_t elsize)
{
	void *p;

	p = calloc(nelem, elsize);
	if (NULL == p)
		eprintf("malloc of %u bytes failed:", nelem * elsize);
	return p;
}

/* allocate new host configuration */
struct host_cfg *new_host_cfg(void)
{
	struct host_cfg *newp;

	newp = (struct host_cfg *)ecalloc(1, sizeof(struct host_cfg));
	newp->next = NULL;
	return newp;
}

/* allocate new state */
struct state *new_state(void)
{
	struct state *newp;

	newp = (struct state *)ecalloc(1, sizeof(struct state));
	newp->next = NULL;
	return newp;
}

/* add state in front (left) of list */
struct state *add_state_l(struct state *listp, struct state *newp)
{
	newp->next = listp;
	return newp;
}

/* add new state to the state list 
 * args: *s		state list
 *       *host		host to be added
 *       *pty		pty to be added
 *
 * return values:	pointer to new state
 */
struct state *add_state(struct state *s, const char *host, int pty)
{
	dprintf(7, "add_state: tty: %d", pty);
	dprintf(7, "add_state: host: %s", host);
	dprintf(7, "add_state: state at: %lx", s);
	s = add_state_l(s, new_state());
	dprintf(7, "add_state: added state at: %lx", s);
	s->host = estrdup(host);
	dprintf(7, "add_state: added host: %s", s->host);
	s->pty = pty;
	s->hashes = 0;
	s->failures = 0;
	s->automated_action = 0;
	s->next_expected_hash = NULL;
	s->maintenance = FALSE;

	return s;
}

/* remove state from the state list */
struct state *rem_state(struct state *state, const char *host, int pty)
{
	struct state *p, *prev;

	prev = NULL;
	for (p = state; p != NULL; p = p->next) {
		dprintf(7, "rem_state: state: %lx", p);
		if (0 == strncmp(host, p->host, MAX_LINE_LENGTH) &&
		    pty == p->pty) {
			if (NULL == prev)
				state = p->next;
			else
				prev->next = p->next;
			free(p);
			return state;
		}
		prev = p;
	}
	/* this happens when the parsing program is started after a pty
	   has been opened and has never been allocated a state. */
	/* TODO: what to do when not in list */

	return state;
}

/* get pointer to the state describing (host, pty)
 * args: *s		current state
 * 	 *h		host configuration
 * 	 pty		pty number
 * 
 * return values:	NULL: nothing found
 * 			pointer to corresponding state on success
 */
struct state *get_state(const struct state *s, const char *h, int pty)
{
	struct state *r = (struct state *)s;

	dprintf(7, "get_state: finding state: %lx, %s, %d", s, h, pty);
	for (; r != NULL; r = r->next) {
		dprintf(7, "get_state: checking state: %lx, %s, %d",
			r, r->host, r->pty);
		if (0 == strncmp(h, r->host, MAX_LINE_LENGTH) && pty == r->pty) {
			dprintf(7, "get_state: found state: %lx", r);
			return r;
		}
	}
	dprintf(7, "get_state: did not find state");
	return NULL;
}

/* get pointer to the host configuration describing the the host 'name' 
 * args: *h		host configuration
 *       *name		hostname to search for
 *
 * return values:	NULL: nothing found
 * 			pointer to host_cfg on success
 */
struct host_cfg *find_host(const struct host_cfg *h, const char *name)
{
	struct host_cfg *t = (struct host_cfg *)h;

	for (; t != NULL; t = t->next) {
		dprintf(7, "find_host: checking host config at '%lx'", t);
		if (0 == strncmp(name, t->name, MAX_LINE_LENGTH))
			return t;
	}
	return NULL;
}

/* check whether we're in maintenance or are going into maintenance 
 * args: *config	configuration data
 *       *s		state data for this host,pty
 *       *hash		hash to check
 *
 * return values	0: pty not in maintenance
 * 			1: pty in maintenance
 *			2: pty now in maintenance
 */
int check_maintenance(struct config *config, struct state *s, const char *hash)
{
	struct host_cfg *h = NULL;

	dprintf(7, "check_maintenance: state at: %lx", s);
	dprintf(9, "check_maintenance: number of failures: %d", s->failures);
	dprintf(9, "check_maintenance: state->maintenance: ", s->maintenance);

	h = find_host(config->host, s->host);
	if (h != NULL) {
		dprintf(7, "check_maintenance: host found: %s", h);
		if (NONE == h->admin_mode) {
			dprintf(8,
				"check_maintenance: no admin mode for this host");
			return 0;
		}
		if (1 == s->maintenance) {
			dprintf(9, "check_maintenance: s->maintenance: %d",
				s->maintenance);
			dprintf(8, "check_maintenance: host in in maintenance");
			return 1;
		}

		if (GLOBAL == h->admin_mode) {
			dprintf(8,
				"check_maintenance: host configured for global admin mode");
			if (1 == cmp_hash(hash, config->admin_digest)) {
				dprintf(8,
					"check_maintenance: host is in maintenance");
				dprintf(9,
					"check_maintenance: setting state->maintenance: ",
					s->maintenance);
				s->maintenance = 1;
				dprintf(9,
					"check_maintenance: set state->maintenance: ",
					s->maintenance);
				return 2;
			} else {
				dprintf(8,
					"check_maintenance: not admin mode!");
			}
		} else {
			dprintf(8,
				"check_maintenance: host configured for host admin mode");
			if (1 == cmp_hash(hash, h->admin_digest)) {
				dprintf(8,
					"check_maintenance: host is in maintenance");
				dprintf(9,
					"check_maintenance: setting state->maintenance: ",
					s->maintenance);
				s->maintenance = 1;
				dprintf(9,
					"check_maintenance: set state->maintenance: ",
					s->maintenance);
				return 2;
			} else {
				dprintf(8,
					"check_maintenance: not admin mode!");
			}
		}
	} else
		dprintf(7, "check_maintenance: no host found");

	dprintf(9,
		"check_maintenance: incrementing number of failures: %d",
		s->failures);
	s->failures++;
	dprintf(9, "check_maintenance: incremented number of failures: %d",
		s->failures);
	return 0;
}

/* compare hash as its logged (ascii) to hash as its stored (binary) 
 * args: *asc_hash	hash in hex notation
 *       *bin_hash	binary hash
 *
 * return values:	0: hashes are not equal (failure)
 * 			1: hashes are equal (success)
 */
int cmp_hash(const char *asc_hash, const unsigned char *bin_hash)
{
	char asc_hash2[HASH_LENGTH];

	if (bin2asc(asc_hash2, bin_hash) < 0)
		return -1;

	dprintf(9, "cmp_hash: hash1 %s hash2 %s", asc_hash, asc_hash2);
	if (0 == strncmp(asc_hash, asc_hash2, HASH_LENGTH)) {
		dprintf(8, "cmp_hash: match found");
		return 1;
	}

	dprintf(8, "cmp_hash: no match found");
	return 0;
}

/* find pointer to cmd_def and to cmd itself for specified hash and hostname:
   1) search config for host with specified hostname
   2) search config->command foreach in  host->allowed_command_set looking for hash
   return current config->command and current command in config->command->command_list
 */
struct cmd_info *find_cmd_info(struct config *config, struct state *state,
			       char *hash)
{
	struct cmd_info *cmd_info;
	struct host_cfg *h;
	char *hostname;

	hostname = state->host;

	cmd_info = NULL;

	h = find_host(config->host, hostname);
	if (NULL == h)
		return NULL;
	dprintf(9, "find_cmd_info with hash %s", hash);
	dprintf(9, "calling find_hash %lx, %lx, %lx", h->allowed_command_set,
		config->command, hash);
	cmd_info =
	    find_hash(h->allowed_command_set, config->command, hash, state);
	if (cmd_info != NULL) {
		dprintf(8, "find_cmd_info: cmd_info found");
		return cmd_info;
	}

	dprintf(8, "find_cmd_info: nothing found");
	return NULL;
}

/* search cmd_def foreach in allowed_command_set looking for hash */
struct cmd_info *find_hash(struct tsw_cmd *allowed_cmd_set,
			   struct cmd_def *cmd_def, char *hash,
			   struct state *state)
{
	struct cmd_info *r;
	struct tsw_cmd *tmp_c;	/* allowed command */
	struct cmd_def *tmp_cd;	/* command definition */

	dprintf(8, "find_hash with hash %s", hash);

	r = NULL;

	for (tmp_c = allowed_cmd_set; tmp_c != NULL; tmp_c = tmp_c->next) {
		for (tmp_cd = cmd_def; tmp_cd != NULL; tmp_cd = tmp_cd->next) {
			if (0 ==
			    strncmp(tmp_c->command, tmp_cd->name,
				    MAX_LINE_LENGTH)) {
				dprintf(9,
					"searching command definition %s (%s)",
					tmp_cd->name, tmp_c->command);
				r = find_hash_helper(tmp_cd, state, hash);
				if (r != NULL)
					return r;
			}
		}
	}
	dprintf(8, "find_hash: nothing found");
	return NULL;
}

/* return > 0 if 'now' is within cmd_def's time specs 
   return 0 if not
 */
int time_ok(struct tsw_time *tsw_time)
{
	time_t clock;
	char tmp[10];		/* hold strftime strings */
	char tmp1[2];
	int weekday;
	int hour;
	int minute;
	int second;
	struct tsw_time *tmp_time;
	struct se_time *se;

	tzset();
	clock = time(NULL);

	/* kind of ugly, but avoids subsequent calls to strftime */
	strftime(tmp, 100, "%w %H %M %S", localtime(&clock));
	tmp1[0] = tmp[0];
	tmp1[1] = '\0';
	weekday = atoi(tmp1);
	tmp1[0] = tmp[2];
	tmp1[1] = tmp[3];
	tmp1[2] = '\0';
	hour = atoi(tmp1);
	tmp1[0] = tmp[5];
	tmp1[1] = tmp[6];
	tmp1[2] = '\0';
	minute = atoi(tmp1);
	tmp1[0] = tmp[8];
	tmp1[1] = tmp[9];
	tmp1[2] = '\0';
	second = atoi(tmp1);

	dprintf(7, "time_ok: localtime %s\n", tmp);
	dprintf(7, "time_ok: weekday %d hour %d minute %d second %d\n",
		weekday, hour, minute, second);
	fflush(stdout);

	for (tmp_time = tsw_time; tmp_time != NULL; tmp_time = tmp_time->next) {
		for (se = tmp_time->se; se != NULL; se = se->next) {
			if (se->weekday[weekday] == SET &&	/* weekday */
			    (hour > se->start.hour ||	/* start time */
			     ((hour == se->start.hour &&
			       minute > se->start.minute) ||
			      (hour == se->start.hour &&
			       minute == se->start.minute &&
			       second >= se->start.second)
			     )
			    ) && (hour < se->end.hour ||	/* end time */
				  ((hour == se->end.hour &&
				    minute < se->end.minute) ||
				   (hour == se->start.hour &&
				    minute == se->end.minute &&
				    second <= se->end.second)
				  )
			    )
			    ) {
				dprintf(8,
					"time_ok: within %d %d %d - %d %d %d",
					se->start.hour, se->start.minute,
					se->start.second, se->end.hour,
					se->end.minute, se->end.second);
				return 1;
			}
		}
	}

	return 0;
}

/* add address of command into state's array for found addresses 
   update number of hashes encountered
 */
int update_state_cmd(struct state *state, struct tsw_cmd *command)
{
	state->cmd_addr = add_cmd_addr_front(state->cmd_addr, new_cmd_addr());
	state->cmd_addr->addr = command;
	state->hashes++;

	return 1;
}

struct cmd_addr *new_cmd_addr(void)
{
	struct cmd_addr *newp;

	newp = (struct cmd_addr *)ecalloc(1, sizeof(struct cmd_addr));
	newp->next = NULL;
	return newp;
}

struct cmd_addr *add_cmd_addr_front(struct cmd_addr *listp,
				    struct cmd_addr *newp)
{
	newp->next = listp;
	return newp;
}

struct se_time *add_se_time_front(struct se_time *listp, struct se_time *newp)
{
	newp->next = listp;
	return newp;
}

struct se_time *new_se_time(void)
{
	struct se_time *newp;

	newp = (struct se_time *)ecalloc(1, sizeof(struct se_time));
	newp->next = NULL;
	return newp;
}

/* check whether command is within list of command_addr 
   returns 0 if not found
 */
int check_cmd_addr(struct cmd_addr *cmd_addr, struct tsw_cmd *command)
{
	struct cmd_addr *tmp;
	for (tmp = cmd_addr; tmp != NULL; tmp = tmp->next)
		if (tmp->addr == command)
			return 1;

	return 0;
}

/* finds cmd to specified hash within specified cmd_def 
   returns it if it's not already listed in the specified state
 */
struct cmd_info *find_hash_helper(struct cmd_def *cmd_def, struct state *state,
				  char *hash)
{
	struct cmd_info *r;
	struct tsw_cmd *tmp_cmd;

	r = NULL;

	/* search for a exact match */
	for (tmp_cmd = cmd_def->command_list; tmp_cmd != NULL;
	     tmp_cmd = tmp_cmd->next) {
		dprintf(10, "checking command %s ", tmp_cmd->command);
		/* check whether this is an already encountered digest */
		if (check_cmd_addr(state->cmd_addr, tmp_cmd)) {
			dprintf(11, "command used before!");
			continue;
		}
		dprintf(11,
			"calling cmp_hash with hash %s and tmp_cmd->digest at %lx",
			hash, tmp_cmd->digest);
		if (cmp_hash(hash, tmp_cmd->digest)) {
			r = ecalloc(1, sizeof(r));
			r->cmd_def = cmd_def;
			r->command = tmp_cmd;
			dprintf(12, "find_hash_helper: hash found");
			return r;
		}
	}
	/* search for an "ANY" command type match */
	dprintf(8, "find_hash_helper: looking for an any command");
	for (tmp_cmd = cmd_def->command_list; tmp_cmd != NULL;
	     tmp_cmd = tmp_cmd->next) {
		if (check_cmd_addr(state->cmd_addr, tmp_cmd)) {
			continue;
		}
		dprintf(8, "find_hash_helper: cmd->mode is %d", tmp_cmd->mode);
		if (tmp_cmd->mode == ANY_CMD) {
			r = ecalloc(1, sizeof(r));
			r->cmd_def = cmd_def;
			r->command = tmp_cmd;
			dprintf(8, "find_hash_helper: an any command found");
			return r;
		}
	}

	return r;
}

/* preprocess the configuration data
 * args: *config		pointer to configuration data
 *
 * return values:		0: success
 * 				>0: failure
 */
int preprocess_config(struct config *config)
{
	int errors = 0;
	if (config != NULL) {
		errors = preprocess_time_spec(config);
		errors += preprocess_syslog_spec(config);
		errors += maint_pass(config);
	} else
		errors++;
	return errors;
}

/* preprocess time specification given in configuration file into
 * a format that is easier to check at runtime
 * args: config		configuration
 *
 * return values:	1: no configuration found
 * 			0: success
 */
int preprocess_time_spec(struct config *config)
{
	struct cmd_def *tmp_c;
	struct host_cfg *tmp_h;
	struct tsw_time *tmp_t;
	int i;

	if (NULL == config)
		return 1;
	dprintf(8, "preprocess_time_spec: processing time");

	/* command definitions */
	for (tmp_c = config->command; tmp_c != NULL; tmp_c = tmp_c->next) {
		dprintf(9,
			"preprocess_time_spec: processing time for command %s",
			tmp_c->name);
		for (tmp_t = tmp_c->time; tmp_t != NULL; tmp_t = tmp_t->next) {
			for (i = 0; i < NR_OF_WEEKDAYS; i++) {
				if (tmp_t->type == PERIODIC
				    && tmp_t->weekday[i] == SET)
					preprocess_time(tmp_t, i);
			}
		}
	}

	/* host definitions (automated time) */
	for (tmp_h = config->host; tmp_h != NULL; tmp_h = tmp_h->next) {
		dprintf(9,
			"preprocess_time_spec: processing time for host %s",
			tmp_h->name);
		for (tmp_t = tmp_h->automated_time; tmp_t != NULL;
		     tmp_t = tmp_t->next) {
			for (i = 0; i < NR_OF_WEEKDAYS; i++) {
				if (tmp_t->type == PERIODIC
				    && tmp_t->weekday[i] == SET)
					preprocess_time(tmp_t, i);
			}
		}
	}

	dprintf(8, "preprocess_time_spec: done with processing time");
	return 0;
}

int preprocess_time(struct tsw_time *time_spec, int day)
{
	struct se_time *se;
	long delta;
	long midnight;
	int d, h, m;

	dprintf(9, "preprocess_time: processing time for day %d", day);

	se = ecalloc(1, sizeof(struct se_time));
	se->start.hour = time_spec->start_hour;
	se->start.minute = time_spec->start_minute;
	se->start.second = 0;
	se->weekday[day] = SET;
	d = day;

	delta = calc_delta(time_spec);
	delta %= WEEK_SECONDS;	/* mod 1 week */

	while (delta > 0) {
		dprintf(10, "preprocess_time: delta %ld", delta);

		midnight = (24 - se->start.hour) * HOUR_SECONDS
		    - se->start.minute * MINUTE_SECONDS - se->start.second;

		add_start(time_spec, d, se);
		if (delta > midnight) {
			add_end(time_spec, d, 23, 59, 59);
			delta -= midnight;
			se->start.hour = 0;
			se->start.minute = 0;
			se->start.second = 0;
			d++;
			d %= 7;
		} else {
			h = floor(delta / (HOUR_SECONDS));
			se->end.hour = se->start.hour + h;
			add_end(time_spec, d, se->end.hour, se->end.minute,
				se->end.second);
			delta -= h * HOUR_SECONDS;

			m = floor(delta / MINUTE_SECONDS);
			se->end.minute = se->start.minute + m;
			delta -= m * MINUTE_SECONDS;

			se->end.second = se->start.second + delta;
			delta = 0;

			/* fixup */
			if (se->end.second > 59) {
				se->end.second -= 60;
				se->end.minute++;
				if (se->end.minute > 59) {
					se->end.minute -= 60;
					se->end.hour++;
				}
			}

		}
	}

	free(se);
	return 0;
}

long calc_delta(struct tsw_time *time_spec)
{
	long delta;		/* in seconds */

	delta = 0;
	if (&time_spec->cont_day != NULL)
		delta += time_spec->cont_day * DAY_SECONDS;
	dprintf(13, "calc_delta %ld (%d %d)", delta, time_spec->cont_day,
		DAY_SECONDS);
	if (&time_spec->cont_hour != NULL)
		delta += time_spec->cont_hour * HOUR_SECONDS;
	dprintf(13, "calc_delta %ld (%d %d)", delta, time_spec->cont_hour,
		HOUR_SECONDS);
	if (&time_spec->cont_minute != NULL)
		delta += time_spec->cont_minute * MINUTE_SECONDS;
	dprintf(13, "calc_delta %ld", delta);
	if (&time_spec->cont_second != NULL)
		delta += time_spec->cont_second;
	dprintf(13, "calc_delta %ld", delta);
	return delta;
}

/* read in interactive password if necessary 
 * args: *config	pointer to configuration data
 *
 * return values:	0: success
 */
int maint_pass(struct config *config)
{
	int g_flag = 0;
	int i;
	struct host_cfg *tmp_h;
	char *p = NULL;
	char *s;
	const char *pr = "Passphrase for ";
	SHA_CTX sha;

	for (tmp_h = config->host; tmp_h != NULL; tmp_h = tmp_h->next) {
		if (tmp_h->admin_mode == INTERACTIVE) {
			dprintf(7,
				"maint_pass: interactive password for %s",
				tmp_h->name);
			s = ecalloc(strlen(tmp_h->name) + strlen(pr) + 2,
				    sizeof(char));
			sprintf(s, "%s%s: ", pr, tmp_h->name);
			dprintf(7, "now reading passphrase");
			while (p == NULL)
				p = read_pass(s);
			dprintf(7, "passphrase is %s", p);
			free(s);
			SHAInit(&sha);
			dprintf(7, "strlen %s is %d", p, (int *)strlen(p));
			SHAUpdate(&sha, (BYTE *) p, strlen(p));
			SHAFinal(tmp_h->admin_digest, &sha);
			/* clear the passphrase */
			for (i = 0; i < strlen(p); i++)
				p[0] = '\0';
			free(p);
		} else if (tmp_h->admin_mode == GLOBAL) {
			dprintf(7, "maint_pass: global password for %s",
				tmp_h->name);
			g_flag++;
		}
	}

	if (g_flag && config->admin_mode == INTERACTIVE) {
		while (p == NULL)
			p = read_pass("Global passphrase: ");
		SHAInit(&sha);
		dprintf(7, "strlen %s is %d", p, (int *)strlen(p));
		SHAUpdate(&sha, (BYTE *) p, strlen(p));
		SHAFinal(config->admin_digest, &sha);
		/* clear the passphrase */
		for (i = 0; i < strlen(p); i++)
			p[0] = '\0';
		free(p);
	}

	return 0;
}

char *read_pass(const char *s)
{
	char *passphrase, *p;
	p = getpass(s);
	passphrase = (char *)calloc(strlen(p) + 1, sizeof(char));
	strncpy(passphrase, p, strlen(p) + 1);
	dprintf(7, "passphrase is %s", passphrase);
	dprintf(8, "passphrase length is %d", strlen(passphrase));
	if (strlen(passphrase) > 0)
		return passphrase;
	return NULL;
}

int automated_action(struct state *state, struct host_cfg *host_cfg)
{
	assert(state != NULL);
	dprintf(2, "automated_action: state %lx", state);
	state->automated_action++;
	dprintf(2, "automated_action: for host %s", host_cfg->name);

	/* are we within an automated action for this state? */
	/* (no locking ...) */
	if (state->automated_action > 1)
		return 1;

	if (time_ok(host_cfg->automated_time)) {
		dprintf(3, "automated_action: within automated time");
		/* let's rock */
		switch (fork()) {
		case -1:
			/* fork failed: TODO do something! */
			return (-1);
		case 0:
			/* child */
			dprintf(4,
				"automated_action: executing command '%s'",
				host_cfg->automated_action);
			system(host_cfg->automated_action);
			log_automated_action(host_cfg);
			exit(0);
		default:
			break;

		}
		return 2;
	} else
		dprintf(3, "automated_action: not within automated time");

	return 0;
}

int log_alert(struct host_cfg *host_cfg)
{
	dprintf(2, "log_alert: for host %s", host_cfg->name);
	openlog(host_cfg->syslog_tag, LOG_CONS | LOG_NDELAY,
		host_cfg->facility);
	syslog(host_cfg->priority, host_cfg->syslog_message);
	closelog();
	return 0;
}

int log_automated_action(struct host_cfg *host_cfg)
{
	dprintf(2, "log_alert: for host %s", host_cfg->name);
	openlog(host_cfg->syslog_tag, LOG_CONS | LOG_NDELAY,
		host_cfg->facility);
	syslog(host_cfg->priority, "automated action executed");
	closelog();
	return 0;
}

int log_ok(struct host_cfg *host_cfg)
{
	dprintf(2, "log_ok: for host %s", host_cfg->name);
	openlog(host_cfg->syslog_tag, LOG_CONS | LOG_NDELAY,
		host_cfg->facility);
	syslog(LOG_DEBUG, "expected command (ok).");
	closelog();
	return 0;
}

int add_start(struct tsw_time *time_spec, int day, struct se_time *se)
{
	dprintf(11, "add_start: day %d %d:%d:%d", day, se->start.hour,
		se->start.minute, se->start.second);
	time_spec->se = add_se_time_front(time_spec->se, new_se_time());
	time_spec->se->start.hour = se->start.hour;
	time_spec->se->start.minute = se->start.minute;
	time_spec->se->start.second = se->start.second;
	time_spec->se->weekday[day] = SET;
	dprintf(12,
		"add_start: time_spec->se %lx day %d %d:%d:%d weekday %d",
		time_spec->se, day, time_spec->se->start.hour,
		time_spec->se->start.minute, time_spec->se->start.second,
		time_spec->se->weekday[day]);
	return 0;
}

int add_end(struct tsw_time *time_spec, int day, int hour, int minute,
	    int second)
{
	dprintf(11, "add_end: day %d %d:%d:%d", day, hour, minute, second);
	assert(time_spec->se->weekday[day] == SET);
	time_spec->se->end.hour = hour;
	time_spec->se->end.minute = minute;
	time_spec->se->end.second = second;
	return 0;
}

/* preprocess the syslog specification from the configuration file into
 * a format that is easier to use at runtime
 * args: *config	pointer to configuration data
 *
 * return values:	0: success
 * 			>0: failure
 */
int preprocess_syslog_spec(struct config *config)
{
	struct host_cfg *tmp_h;

	for (tmp_h = config->host; tmp_h != NULL; tmp_h = tmp_h->next) {

		dprintf(1, "preprocess_syslog_spec: %s", tmp_h->name);
		/* facility munging */
		if (strncasecmp(tmp_h->syslog_facility, "user", MAX_LINE_LENGTH)
		    == 0) {
			tmp_h->facility = LOG_USER;
		} else if (strncasecmp(tmp_h->syslog_facility, "local", 5) == 0) {
			switch (tmp_h->syslog_facility[5]) {
			case '0':
				tmp_h->facility = LOG_LOCAL0;
			case '1':
				tmp_h->facility = LOG_LOCAL1;
			case '2':
				tmp_h->facility = LOG_LOCAL2;
			case '3':
				tmp_h->facility = LOG_LOCAL3;
			case '4':
				tmp_h->facility = LOG_LOCAL4;
			case '5':
				tmp_h->facility = LOG_LOCAL5;
			case '6':
				tmp_h->facility = LOG_LOCAL6;
			case '7':
				tmp_h->facility = LOG_LOCAL7;
			default:
				tmp_h->facility = LOG_DAEMON;
			}

		} else {
			tmp_h->facility = LOG_DAEMON;
		}

		/* priority munging */
		dprintf(1, "preprocess_syslog_spec: priority munging");
		if (strncasecmp
		    (tmp_h->syslog_priority, "debug", MAX_LINE_LENGTH) == 0) {
			dprintf(1, "priority: debug %d", LOG_DEBUG);
			tmp_h->priority = LOG_DEBUG;
		} else
		    if (strncasecmp
			(tmp_h->syslog_priority, "notice",
			 MAX_LINE_LENGTH) == 0) {
			dprintf(1, "priority: error %d", LOG_NOTICE);
			tmp_h->priority = LOG_NOTICE;
		} else
		    if (strncasecmp
			(tmp_h->syslog_priority, "warning",
			 MAX_LINE_LENGTH) == 0) {
			dprintf(1, "priority: error %d", LOG_WARNING);
			tmp_h->priority = LOG_WARNING;
		} else
		    if (strncasecmp
			(tmp_h->syslog_priority, "error",
			 MAX_LINE_LENGTH) == 0) {
			tmp_h->priority = LOG_ERR;
			dprintf(1, "priority: error %d", LOG_ERR);
		} else
		    if (strncasecmp
			(tmp_h->syslog_priority, "critical",
			 MAX_LINE_LENGTH) == 0) {
			dprintf(1, "priority: error %d", LOG_CRIT);
			tmp_h->priority = LOG_CRIT;
		} else
		    if (strncasecmp
			(tmp_h->syslog_priority, "alert",
			 MAX_LINE_LENGTH) == 0) {
			dprintf(1, "priority: error %d", LOG_ALERT);
			tmp_h->priority = LOG_ALERT;
		} else
		    if (strncasecmp
			(tmp_h->syslog_priority, "emergency",
			 MAX_LINE_LENGTH) == 0) {
			dprintf(1, "priority: emergency %d", LOG_EMERG);
			tmp_h->priority = LOG_EMERG;
		} else {
			dprintf(1, "priority: crit %d", LOG_CRIT);
			tmp_h->priority = LOG_CRIT;
		}

	}

	return 0;
}

struct cmd_line_t *cmd_parse(int argc, char **argv)
{
	int c;
	extern char *optarg;
	extern int optind;
	struct cmd_line_t *cmd_line_opts;

	dprintf(2, "cmd_parse: parsing command line arguments");
	cmd_line_opts =
	    (struct cmd_line_t *)calloc(1, sizeof(struct cmd_line_t));

	while ((c = getopt(argc, argv, "c:d:sq")) != -1) {
		switch (c) {
		case 'c':
			cmd_line_opts->file = estrdup(optarg);
			dprintf(3, "cmd_parse: file is %s",
				cmd_line_opts->file);
			break;
		case 'd':
			cmd_line_opts->debug = atoi(optarg);
			dprintf(3, "cmd_parse: debug is %d",
				cmd_line_opts->debug);
			break;
		case 's':
			cmd_line_opts->s = 1;
			dprintf(3, "cmd_parse: s is %d", cmd_line_opts->s);
			break;
		case 'q':
			cmd_line_opts->q = 1;
			dprintf(3, "cmd_parse: q is %d", cmd_line_opts->q);
			break;
		}
	}
	for (; optind < argc; optind++) {
		printf("extra argument: %s\n", argv[optind]);
	}

	return cmd_line_opts;
}

void usage(void)
{
	printf("%s\n", progname());
	printf("usage: %s -c file [ -d int ] [-q]\n", progname());
	printf("       %s -s [ -d int ] [-q]\n\n", progname());
	printf(" -c file\tconfiguration file\n");
	printf(" -s\t\tinvokes SHA1 password generation\n");
	printf(" -d int\t\tprint debug information (1 < int < 16)\n");
	printf(" -q\t\tquiet output\n");
}

/* find a host name entry for the line received
 * args: *line		line to check
 *       length		line length
 *       *host		to store the hostname
 *       dest_length	maximal length for hostname
 *
 * return values:	-1: no host found
 * 			0:  host found
 */
int find_host_in_line(const char *line, int length, char *host, int dest_length)
{
	char *tmp_host;
	tmp_host = (char *)ecalloc(length, sizeof(char));
	tmp_host[0] = '\0';

	if (sscanf(line, FMT_MATCH_HOST, tmp_host) == 0) {
		dprintf(8, "find_host_in_line: no host found in line %s!",
			line);
		host[0] = '\0';
		return -1;
	}
	snprintf(host, dest_length, "%s", tmp_host);
	dprintf(8, "find_host: host '%s' found.", host);
	return 0;
}

/* find "tty= line= and hash" pattern in line given
 * args: *line		line to search
 *       *hash		to store the hash found
 *       hash_length	maximal hash length to be stored
 *       *tty		to store the tty found
 *	 *linenumber	to store the line number found
 *
 * return values:	-2: could position in line
 * 			-1: did not find pattern
 * 			0:  match successful
 */
int find_tsw_in_line(const char *line, char *hash, int hash_length, int *tty,
		     long *linenumber)
{
	char *tmp_position;
	char block1[MAX_LINE_LENGTH];
	char block2[MAX_LINE_LENGTH];
	char block3[MAX_LINE_LENGTH];
	char block4[MAX_LINE_LENGTH];
	char block5[MAX_LINE_LENGTH];
	int tmp_tty;
	int tmp_linenumber;

	dprintf(8, "find_tsw_in_line");
	dprintf(9, "find_tsw_in_line: line '%s'", line);
	dprintf(9, "find_tsw_in_line: format '%s'", FMT_MATCH_TSW);
	tmp_position = strstr(line, " tty=");
	if (tmp_position == NULL) {
		return -2;
	}
	dprintf(9, "find_tsw_in_line: tmp_position '%s'", tmp_position);
	tmp_position -= 5;	/* position to 'tsw:' */
	dprintf(9, "find_tsw_in_line: tmp_position '%s'", tmp_position);

	if (sscanf
	    (tmp_position, FMT_MATCH_TSW, &tmp_tty, &tmp_linenumber,
	     block1, block2, block3, block4, block5) == 0) {
		dprintf(9, "find_tsw_in_line: sscanf not successful");
		return -1;
	}
	block1[8] = '\0';
	block2[8] = '\0';
	block3[8] = '\0';
	block4[8] = '\0';
	block5[8] = '\0';
	dprintf(9,
		"find_tsw_in_line: tty '%d', line '%d', hash '%s %s %s %s %s'",
		tmp_tty, tmp_linenumber, block1, block2, block3, block4,
		block5);

	/* TODO: check validity */

	*tty = tmp_tty;
	*linenumber = tmp_linenumber;
	snprintf(hash, hash_length, "%s %s %s %s %s", block1, block2,
		 block3, block4, block5);
	hash[HASH_LENGTH] = '\0';
	return 0;
}

int gen_pass(void)
{
	char *p = NULL;
	const char *pr = "Passphrase: ";
	SHA_CTX sha;
	unsigned char bin_hash[SHA1_HASH_SIZE];
	int i;
	char asc_hash[HASH_LENGTH];

	while (p == NULL)
		p = read_pass(pr);
	SHAInit(&sha);
	SHAUpdate(&sha, (BYTE *) p, strlen(p));
	SHAFinal(bin_hash, &sha);
	/* clear the passphrase */
	for (i = 0; i < strlen(p); i++)
		p[0] = '\0';
	free(p);

	if (bin2asc(asc_hash, bin_hash) < 0)
		return -1;
	printf("%s\n", asc_hash);
	return 0;
}

/* read in configuration data 
 * args: *cmd_line_opts	command line arguments
 *
 * return values:	 0: successfully parsed file
 * 			-1: parsing failed
 */
int read_config(char *filename)
{
	extern FILE *yyin;
	extern int yyparse();
	int rc;

	yyin = fopen(filename, "r");
	if (yyin == NULL) {
		printf("error: could not open file %s", filename);
		return -2;
	}
	rc = yyparse();
	fclose(yyin);
	if (rc == 0)
		return 0;
	return -1;
}

/* 
 * open logfile and read its tail
 * args: *c		configuration
 * 
 * return values	0 on success
 * 			-1 close on logfile failed
 */
int tail_on_file(struct config *c)
{
	struct file_spec s_file;

	dprintf(5, "tail_on_file: state is at %lx\n", state);

	if (open_log(&s_file) < 0) {
		perror("tail_on_file: open_log failed.\n");
		return -1;
	}
	if (tail(s_file) < 0) {
		perror("tail_on_file: tail failed.\n");
		return -3;
	}
	if (close(s_file.fd) == 0)
		return 0;

	return -3;
}

/* 
 * open file, store fd, and stat and store its size 
 * args: *s_file 	used to store file descriptor 
 * 			and file size
 * return values:	0 on success
 * 			-1 on file open failure
 * 			-2 on file stat failure
 */
int open_log(struct file_spec *s_file)
{
	struct stat *stats;
	stats = (struct stat *)ecalloc(1, sizeof(struct stat));

	s_file->fd = open(config->logfile, O_RDONLY);
	if (s_file->fd < 0) {
		perror("open_log: cannot open file");
		return -1;
	}
	if (fstat(s_file->fd, stats) < 0) {
		s_file->fd = -1;
		perror("open_log: cannot get stats");
		return -2;
	}
	s_file->size = stats->st_size;

	return 0;
}

/* 
 * tail on the file
 * args: *s_file	used to store file descriptor
 * 			and file size
 * 
 * return values	0 on success (never reached, though)
 * 			-1 stat failed
 */
int tail(struct file_spec s_file)
{
	struct stat *stats;

	stats = (struct stat *)ecalloc(1, sizeof(struct stat));

	lseek(s_file.fd, 0, SEEK_END);

	dprintf(5, "tail: state is at %lx\n", state);

	while (1) {

		if (fstat(s_file.fd, stats) < 0) {
			s_file.fd = -1;
			perror("cannot get stats");
			return -1;
		}

		/* file truncated */
		if (stats->st_size < s_file.size) {
			dprintf(1, "file was truncated");
			fflush(stdout);
			lseek(s_file.fd, stats->st_size, SEEK_SET);
			s_file.size = stats->st_size;
		}

		if (stats->st_size == s_file.size) {
			dprintf(2, "sleeping");
			fflush(stdout);
			sleep(1);
			dprintf(2, "slept");
		} else {
			dprintf(2, "dumping");
			fflush(stdout);
			s_file.size +=
			    dump_remainder(s_file.fd,
					   stats->st_size - s_file.size);
		}

	}
	return 0;
}

/* read and process n_bytes of file starting at the current
 * position in fd
 * args: fd		file descriptor to read on
 *       n_bytes	bytes to read
 *
 * return the number of bytes read from the file.  
 */
long dump_remainder(int fd, off_t n_bytes)
{
	char buffer[BUFSIZ];
	int bytes_read;
	long n_written;
	off_t n_remaining = n_bytes;

	n_written = 0;

	dprintf(5, "dump_remainder: state is at %lx\n", state);

	while (1) {
		long n = MIN(n_remaining, (off_t) BUFSIZ);
		bytes_read = read(fd, buffer, n);
		if (bytes_read <= 0)
			break;

		if (proc_buf(buffer, bytes_read) != 0)
			printf("error processing buffer %s\n", buffer);

		dprintf(8, "bytes read: %d", bytes_read);
		n_remaining -= bytes_read;
		dprintf(8, "n_remaining: %d", n_remaining);
		n_written += bytes_read;
	}

	return n_written;
}

/* split buffer into individual lines
 * process the lines as we encounter them 
 * args: *buf		buffer to process
 *       bytes_read	length of the buffer
 *
 * return values:	<0 on failure
 * 			0 on success
 */

int proc_buf(char *buf, int bytes_read)
{
	int i = 0;
	int j = 0;
	char line[BUFSIZ];
	buf[bytes_read] = '\0';

	dprintf(5, "proc_buf: state is at %lx", state);
	dprintf(8, "processing buffer:\n%s", buf);

	while (i < bytes_read) {
		line[j++] = buf[i];
		if (buf[i] == '\n' || i == bytes_read - 1) {
			line[j - 1] = '\0';
			dprintf(8, "got a line: %s", line);
			proc_line(line, j);
			j = 0;
		}
		i++;
	}
	return 0;
}

/* sort out the lines we're interested in
 * args: *line		line to check
 *       length		length of line
 *
 * return value		0: skipped line (line too short 
 * 				for pty hash)
 * 			1: accepted line for processing
 */
int proc_line(char *line, int length)
{

	dprintf(5, "proc_line: state is at %lx", state);

	if (length < MIN_DIGEST_LINE_LENGTH) {
		dprintf(7, "proc_line: short line");
		return 0;
	}

	if (pty_hash(line, length) > 0)
		dprintf(7, "proc_line: pty hash");

	return 1;
}

/* check whether this is a currently expected hash 
 *  - get the state of this (host, pty) from the state table
 *  - check whether we're in maintenance or going into it now
 * args: *state		pointer to current state
 *       *hash		hash to check
 *       *host		host to check
 *       pty		pty in question
 * 
 * return values	0: digest invalid
 */
int check_digest_validity(const char *hash, const char *host, int pty)
{
	struct state *s = NULL;

	dprintf(6, "check_digest: checking digest validity");

	s = get_state(state, host, pty);
	if (s == NULL) {
		dprintf(7, "check_digest: corresponding state not found");
		/* this happens when the parsing program is started
		   after a pty has been opened and has never been
		   allocated a state */
		state = add_state(state, host, pty);
		s = state;
		/* TODO: what to do when not in list: log something? */
	}
	dprintf(7, "check_digest: corresponding state found");
	dprintf(8, "check_digest: host %s, pty: %d, hashes: %d, failures: %d",
		s->host, s->pty, s->hashes, s->failures);

	if (check_maintenance(config, s, hash) > 0) {
		dprintf(8, "check_digest: check_maintenance  > 0");
		return 1;
	}

	dprintf(7, "check_digest: alarm!");
	return 0;
}

/* search for a hash and a tty check validity if found 
 * args: *line		line to check
 *       length		line length
 *
 * return values	-1: no host found on this line
 * 			-2: host not in configuration
 * 			 0: no SID hash found in line
 * 			 1: SID hash found in line
 * 			 2: opening or terminating SID hash found in line
 *
 */
int pty_hash(const char *line, int length)
{
	char host[MAX_LINE_LENGTH];
	char hash[HASH_LENGTH];
	int tty;
	long ln;		/* linenumber */
	int rt = -1;

	dprintf(5, "pty_hash: state is at %lx", state);
	dprintf(8, "pty_hash");

	rt = check_host(line, length, host, MAX_LINE_LENGTH);
	if (0 != rt)
		return rt;

	if (find_tsw_in_line(line, hash, HASH_LENGTH, &tty, &ln) == 0) {

		dprintf(8, "pty_hash: hash: %s, host: %s, tty: %d, line: %ld",
			hash, host, tty, ln);

		if (0 != fix_open_close(hash, host, tty, ln)) {
			return 2;
		}

		(void *)chk_hash(hash, host, tty);
		return 1;
	} else {
		dprintf(8, "pty_open: find_tsw_in_line failed!");
	}
	return 0;
}

/* wrapper for reading the configuration settings
 * args:	*filename	of the config file
 *
 * return values:	 0: success
 * 			-1: parsing failed
 */
int config_parse(char *filename)
{
	if (read_config(filename) < 0) {
		printf("error: parsing configuration file failed.");
		return -1;
	}
	return 0;
}

/* check correctness of supplied command line args
 * args:	*cmd_line_opts	pointer to the collected args
 *
 * return value:	number of errors (0 on success)
 */
int cmdargs_check(struct cmd_line_t *cmd_line_opts)
{
	int errors = 0;
	if (cmd_line_opts->s != 1 && cmd_line_opts->file == NULL) {
		printf("error: no configuration file specified\n");
		errors++;
	}
	/* TODO: do some further checks */
	return errors;
}

/* fork to background unless started in debug mode
 * args: cmd_line_opts	pointer to command line args
 *
 * return values:	 1: no fork (debug mode)
 * 			-1: fork failed
 */
int become_daemon(const struct cmd_line_t *cmd_line_opts)
{
	if (cmd_line_opts->debug == 0) {
		switch (fork()) {
		case -1:
			return (-1);
		case 0:
			if (cmd_line_opts->q == 0)
				printf("%s started.\n", progname());
			break;
		default:
			exit(0);
		}
	}
	return 1;
}

/* convert binary hash to ascii form 
 * args: *asc_hash	target for ascii hash
 *       *bin_hash	binary hash
 *
 * return values	-1: conversion failed
 * 			 0: success
 */
int bin2asc(char *asc_hash, const unsigned char *bin_hash)
{
	int rc = 0;
	rc = snprintf(asc_hash, HASH_LENGTH,
		      "%02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x",
		      bin_hash[0], bin_hash[1], bin_hash[2], bin_hash[3],
		      bin_hash[4], bin_hash[5], bin_hash[6], bin_hash[7],
		      bin_hash[8], bin_hash[9], bin_hash[10], bin_hash[11],
		      bin_hash[12], bin_hash[13], bin_hash[14],
		      bin_hash[15], bin_hash[16], bin_hash[17],
		      bin_hash[18], bin_hash[19]);
	dprintf(9, "bin2ask: hash1 %s, snprintf returned %d", asc_hash, rc);
	if (rc == HASH_LENGTH - 1) {	/* HASH_LENGTH includes '\0' */
		return 0;
	}
	return -1;
}

/*
 * check if this is the opening or terminating hash
 * if so, deallocate state
 * args: *hash		hash to check
 *       *host		host to check
 *       tty		tty to check
 *       ln		linenumber to check
 *
 * return values	0: no open/close hash
 * 			1: open hash
 * 			2: close hash
 */
int fix_open_close(const char *hash, const char *host, int tty, int ln)
{
	struct state *t;

	if (strncmp(hash, TSW_OPEN_CLOSE, strlen(TSW_OPEN_CLOSE)) == 0) {
		dprintf(9, "fix_open_close: opening/terminating hash ...");
		if (ln == 0) {
			t = get_state(state, host, tty);
			if (t != NULL) {
				dprintf(10,
					"fix_open_close: opening hash with existing state. Deallocating ...");
				state = rem_state(state, host, tty);
				dprintf(10,
					"fix_open_close: ... state deallocated.");
			}
			return 1;
		} else {
			t = get_state(state, host, tty);
			if (t != NULL) {
				dprintf(10,
					"fix_open_close: closing hash with existing state. Deallocating ...");
				state = rem_state(state, host, tty);
				dprintf(10,
					"fix_open_close: ... state deallocated.");
			}
			return 2;
		}
	}
	return 0;
}

/* checks whether a host is found in the line and 
 * whether we've a configuration for this host 
 * args:	*line		line to check
 * 		*host		store the host here if found
 * 		line_length	line length
 * 		host_length	max host length to store
 *
 * return values:		-1: no host found
 * 				-2: host found, no config for this host
 * 				0: host found, config found */
int check_host(const char *line, int line_length, char *host, int host_length)
{
	if (find_host_in_line(line, line_length, host, host_length) < 0) {
		/* TODO: what do we do for lines that have no host? */
		dprintf(1, "check_host: no host found!");
		return -1;
	}

	/* check if we have an entry for this host in the config file */
	if (find_host(config->host, host) == NULL) {
		dprintf(1, "check_host: no configuration for host '%s' found.",
			host);
		return -2;
	}
	return 0;
}

/* check the validity of the hash in the state table
 * args:	*hash		hash to check
 * 		*host		host to check
 * 		tty		tty where the hash appeared
 *
 * return values:		0 on success */
/* FIXME: there no error handling. are there error cases or should this be 
 * void? */
int chk_hash(const char *hash, const char *host, int tty)
{
	struct host_cfg *host_cfg_ptr;
	/* 
	 * allocate state if necessary - state might not be allocated
	 * yet (e.g. pty session exists before this runs)
	 */
	if (state == NULL) {
		dprintf(9,
			"pty_hash: state not yet allocated. Allocating one ...");
		state = add_state(state, host, tty);
		dprintf(9, "pty_hash: state allocated.");
		dprintf(10, "pty_hash: state failures: %d", state->failures);
	}

	if (check_digest_validity(hash, host, tty)) {
		dprintf(1, "pty_hash: hash ok");
		log_ok(find_host(config->host, host));
	} else {
		dprintf(1, "pty_hash: hash not ok");
		host_cfg_ptr = find_host(config->host, host);
		log_alert(host_cfg_ptr);
		/* execute automated action if number of allowed attempts is exceeded */
		if (state->failures >= host_cfg_ptr->max_failures) {
			dprintf(1, "pty_hash: spawning automated action for %s tty %d", host, tty);
			automated_action(get_state(state, host, tty),
					 host_cfg_ptr);
		}
	}
	return 0;
}

/* indent -kr -i8 -ts8 -sob -l80 -ss -ncs */
